@@ -1,20 +1,19 @@
+{-# LANGUAGE TypeFamilies #-}
+{-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE FunctionalDependencies #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE CPP #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE NoImplicitPrelude #-}
+{-# LANGUAGE BangPatterns #-}
 module RIO.Prelude.Logger
-  ( -- * Standard logging functions
-    logDebug
-  , logInfo
-  , logWarn
-  , logError
-  , logOther
-    -- * Running with logging
-  , withLogFunc
+  ( -- ** Running with logging
+    withLogFunc
   , newLogFunc
   , LogFunc
   , HasLogFunc (..)
   , logOptionsHandle
-    -- ** Log options
+    -- *** Log options
   , LogOptions
   , setLogMinLevel
   , setLogMinLevelIO
@@ -24,30 +23,50 @@ module RIO.Prelude.Logger
   , setLogUseTime
   , setLogUseColor
   , setLogUseLoc
-    -- * Advanced logging functions
-    -- ** Sticky logging
+  , setLogFormat
+    -- ** Standard logging functions
+  , logDebug
+  , logInfo
+  , logWarn
+  , logError
+  , logOther
+    -- ** Advanced logging functions
+    -- *** Sticky logging
   , logSticky
   , logStickyDone
-    -- ** With source
+    -- *** With source
+    --
+    -- $withSource
   , logDebugS
   , logInfoS
   , logWarnS
   , logErrorS
   , logOtherS
-    -- ** Generic log function
+    -- *** Generic log function
   , logGeneric
-    -- * Advanced running functions
+    -- ** Advanced running functions
   , mkLogFunc
   , logOptionsMemory
-    -- * Data types
+    -- ** Data types
   , LogLevel (..)
   , LogSource
   , CallStack
-    -- * Convenience functions
+    -- ** Convenience functions
   , displayCallStack
   , noLogging
-    -- * Accessors
+    -- ** Accessors
   , logFuncUseColorL
+    -- * Type-generic logger
+    -- $type-generic-intro
+  , glog
+  , GLogFunc
+  , gLogFuncClassic
+  , mkGLogFunc
+  , contramapMaybeGLogFunc
+  , contramapGLogFunc
+  , HasGLogFunc(..)
+  , HasLogLevel(..)
+  , HasLogSource(..)
   ) where
 
 import RIO.Prelude.Reexports hiding ((<>))
@@ -60,6 +79,7 @@ import Control.Monad.IO.Class (MonadIO, liftIO)
 import GHC.Stack (HasCallStack, CallStack, SrcLoc (..), getCallStack, callStack)
 import Data.Time
 import qualified Data.Text.IO as TIO
+import Data.Bits
 import Data.ByteString.Builder (toLazyByteString, char7, byteString, hPutBuilder)
 import Data.ByteString.Builder.Extra (flush)
 import           GHC.IO.Handle.Internals         (wantWritableHandle)
@@ -70,6 +90,10 @@ import           System.IO                  (localeEncoding)
 import           GHC.Foreign                (peekCString, withCString)
 import Data.Semigroup (Semigroup (..))
 
+#if MIN_VERSION_base(4,12,0)
+import Data.Functor.Contravariant
+#endif
+
 -- | The log level of a message.
 --
 -- @since 0.0.0.0
@@ -182,6 +206,17 @@ logOther
   -> m ()
 logOther = logGeneric "" . LevelOther
 
+-- $withSource
+--
+-- There is a set of logging functions that take an extra 'LogSource'
+-- argument to provide context, typically detailing what part of an
+-- application the message comes from.
+--
+-- For example, in verbose mode, @infoLogS "database" "connected"@ will
+-- result in
+--
+-- > [info] (database) connected
+
 -- | Log a debug level message with the given source.
 --
 -- @since 0.0.0.0
@@ -285,6 +320,7 @@ logOptionsMemory = do
         , logUseTime = False
         , logUseColor = False
         , logUseLoc = False
+        , logFormat = id
         , logSend = \new -> atomicModifyIORef' ref $ \old -> (old <> new, ())
         }
   return (ref, options)
@@ -322,6 +358,7 @@ logOptionsHandle handle' verbose = liftIO $ do
     , logUseColor = verbose && terminal
 #endif
     , logUseLoc = verbose
+    , logFormat = id
     , logSend = \builder ->
         if useUtf8 && unicode
           then hPutBuilder handle' (builder <> flush)
@@ -355,16 +392,16 @@ getCanUseUnicode = do
 -- Intended for use if you want to deal with the teardown of 'LogFunc' yourself,
 -- otherwise prefer the 'withLogFunc' function instead.
 --
---  @since  0.1.3.0
+--  @since 0.1.3.0
 newLogFunc :: (MonadIO n, MonadIO m) => LogOptions -> n (LogFunc, m ())
 newLogFunc options =
   if logTerminal options then do
-    var <- newMVar mempty
+    var <- newMVar (mempty,0)
     return (LogFunc
              { unLogFunc = stickyImpl var options (simpleLogFunc options)
              , lfOptions = Just options
              }
-           , do state <- takeMVar var
+           , do (state,_) <- takeMVar var
                 unless (B.null state) (liftIO $ logSend options "\n")
            )
   else
@@ -383,7 +420,7 @@ newLogFunc options =
 -- let isVerbose = False -- get from the command line instead
 -- logOptions' <- logOptionsHandle stderr isVerbose
 -- let logOptions = setLogUseTime True logOptions'
--- withLogFunc logOptions $ \lf -> do
+-- withLogFunc logOptions $ \\lf -> do
 --   let app = App -- application specific environment
 --         { appLogFunc = lf
 --         , appOtherStuff = ...
@@ -423,6 +460,7 @@ data LogOptions = LogOptions
   , logUseTime :: !Bool
   , logUseColor :: !Bool
   , logUseLoc :: !Bool
+  , logFormat :: !(Utf8Builder -> Utf8Builder)
   , logSend :: !(Builder -> IO ())
   }
 
@@ -474,7 +512,7 @@ setLogTerminal t options = options { logTerminal = t }
 
 -- | Include the time when printing log messages.
 --
--- Default: true in debug mode, false otherwise.
+-- Default: `True` in debug mode, `False` otherwise.
 --
 -- @since 0.0.0.0
 setLogUseTime :: Bool -> LogOptions -> LogOptions
@@ -482,7 +520,7 @@ setLogUseTime t options = options { logUseTime = t }
 
 -- | Use ANSI color codes in the log output.
 --
--- Default: true if in verbose mode /and/ the 'Handle' is a terminal device.
+-- Default: `True` if in verbose mode /and/ the 'Handle' is a terminal device.
 --
 -- @since 0.0.0.0
 setLogUseColor :: Bool -> LogOptions -> LogOptions
@@ -490,14 +528,22 @@ setLogUseColor c options = options { logUseColor = c }
 
 -- | Use code location in the log output.
 --
--- Default: true if in verbose mode, false otherwise.
+-- Default: `True` if in verbose mode, `False` otherwise.
 --
 -- @since 0.1.2.0
 setLogUseLoc :: Bool -> LogOptions -> LogOptions
 setLogUseLoc l options = options { logUseLoc = l }
 
+-- | Set format method for messages
+--
+-- Default: `id`
+--
+-- @since 0.1.13.0
+setLogFormat :: (Utf8Builder -> Utf8Builder) -> LogOptions -> LogOptions
+setLogFormat f options = options { logFormat = f }
+
 simpleLogFunc :: LogOptions -> CallStack -> LogSource -> LogLevel -> Utf8Builder -> IO ()
-simpleLogFunc lo cs _src level msg = do
+simpleLogFunc lo cs src level msg = do
     logLevel   <- logMinLevel lo
     logVerbose <- logVerboseFormat lo
 
@@ -507,7 +553,8 @@ simpleLogFunc lo cs _src level msg = do
         timestamp <>
         getLevel logVerbose <>
         ansi reset <>
-        msg <>
+        getSource <>
+        logFormat lo msg <>
         getLoc <>
         ansi reset <>
         "\n"
@@ -549,6 +596,11 @@ simpleLogFunc lo cs _src level msg = do
              "] "
      | otherwise = mempty
 
+   getSource :: Utf8Builder
+   getSource = case src of
+     "" -> ""
+     _  -> "(" <> display src <> ") "
+
    getLoc :: Utf8Builder
    getLoc
      | logUseLoc lo = ansi setBlack <> "\n@(" <> displayCallStack cs <> ")"
@@ -579,12 +631,12 @@ timestampLength =
   length (formatTime defaultTimeLocale "%F %T.000000" (UTCTime (ModifiedJulianDay 0) 0))
 
 stickyImpl
-    :: MVar ByteString -> LogOptions
+    :: MVar (ByteString,Int) -> LogOptions
     -> (CallStack -> LogSource -> LogLevel -> Utf8Builder -> IO ())
     -> CallStack -> LogSource -> LogLevel -> Utf8Builder -> IO ()
-stickyImpl ref lo logFunc loc src level msgOrig = modifyMVar_ ref $ \sticky -> do
+stickyImpl ref lo logFunc loc src level msgOrig = modifyMVar_ ref $ \(sticky,stickyLen) -> do
   let backSpaceChar = '\8'
-      repeating = mconcat . replicate (B.length sticky) . char7
+      repeating = mconcat . replicate stickyLen . char7
       clear = logSend lo
         (repeating backSpaceChar <>
         repeating ' ' <>
@@ -596,19 +648,37 @@ stickyImpl ref lo logFunc loc src level msgOrig = modifyMVar_ ref $ \sticky -> d
     LevelOther "sticky-done" -> do
       clear
       logFunc loc src LevelInfo msgOrig
-      return mempty
+      return (mempty,0)
     LevelOther "sticky" -> do
       clear
       let bs = toStrictBytes $ toLazyByteString $ getUtf8Builder msgOrig
       logSend lo (byteString bs <> flush)
-      return bs
+      return (bs, utf8CharacterCount bs)
     _
       | level >= logLevel -> do
           clear
           logFunc loc src level msgOrig
           unless (B.null sticky) $ logSend lo (byteString sticky <> flush)
-          return sticky
-      | otherwise -> return sticky
+          return (sticky,stickyLen)
+      | otherwise -> return (sticky,stickyLen)
+
+-- | The number of Unicode characters in a UTF-8 encoded byte string,
+-- excluding ANSI CSI sequences.
+utf8CharacterCount :: ByteString -> Int
+utf8CharacterCount = go 0
+  where
+    go !n bs = case B.uncons bs of
+        Nothing -> n
+        Just (c,bs)
+            | c .&. 0xC0 == 0x80 -> go n bs            -- UTF-8 continuation
+            | c == 0x1B          -> go n $ dropCSI bs  -- ANSI escape
+            | otherwise          -> go (n+1) bs
+
+    dropCSI bs = case B.uncons bs of
+        Just (0x5B,bs2) -> B.drop 1 $ B.dropWhile isSequenceByte bs2
+        _               -> bs
+
+    isSequenceByte c = c >= 0x20 && c <= 0x3F
 
 -- | Is the log func configured to use color output?
 --
@@ -627,3 +697,190 @@ logFuncUseColorL = logFuncL.to (maybe False logUseColor . lfOptions)
 -- @since 0.1.5.0
 noLogging :: (HasLogFunc env, MonadReader env m) => m a -> m a
 noLogging = local (set logFuncL mempty)
+
+--------------------------------------------------------------------------------
+--
+-- $type-generic-intro
+--
+-- When logging takes on a more semantic meaning and the logs need to
+-- be digested, acted upon, translated or serialized upstream (to
+-- e.g. a JSON logging server), we have 'GLogFunc' (as in "generic log
+-- function"), and is accessed via 'HasGLogFunc'.
+--
+-- There is only one function to log in this system: the 'glog'
+-- function, which can log any message. You determine the log levels
+-- or severity of messages when needed.
+--
+-- Using 'RIO.Prelude.mapRIO' and 'contramapGLogFunc' (or
+-- 'contramapMaybeGLogFunc'), you can build hierarchies of loggers.
+--
+-- Example:
+--
+-- @
+-- import RIO
+--
+-- data DatabaseMsg = Connected String | Query String | Disconnected deriving Show
+-- data WebMsg = Request String | Error String | DatabaseMsg DatabaseMsg deriving Show
+-- data AppMsg = InitMsg String | WebMsg WebMsg deriving Show
+--
+-- main :: IO ()
+-- main =
+--   runRIO
+--     (mkGLogFunc (\stack msg -> print msg))
+--     (do glog (InitMsg "Ready to go!")
+--         runWeb
+--           (do glog (Request "/foo")
+--               runDB (do glog (Connected "127.0.0.1")
+--                         glog (Query "SELECT 1"))
+--               glog (Error "Oh noes!")))
+--
+-- runDB :: RIO (GLogFunc DatabaseMsg) () -> RIO (GLogFunc WebMsg) ()
+-- runDB = mapRIO (contramapGLogFunc DatabaseMsg)
+--
+-- runWeb :: RIO (GLogFunc WebMsg) () -> RIO (GLogFunc AppMsg) ()
+-- runWeb = mapRIO (contramapGLogFunc WebMsg)
+-- @
+--
+-- If we instead decided that we only wanted to log database queries,
+-- and not bother the upstream with connect/disconnect messages, we
+-- could simplify the constructor to @DatabaseQuery String@:
+--
+-- @
+-- data WebMsg = Request String | Error String | DatabaseQuery String deriving Show
+-- @
+--
+-- And then @runDB@ could use 'contramapMaybeGLogFunc' to parse only queries:
+--
+-- @
+-- runDB =
+--   mapRIO
+--     (contramapMaybeGLogFunc
+--        (\msg ->
+--           case msg of
+--             Query string -> pure (DatabaseQuery string)
+--             _ -> Nothing))
+-- @
+--
+-- This way, upstream only has to care about queries and not
+-- connect/disconnect constructors.
+
+-- | An app is capable of generic logging if it implements this.
+--
+-- @since 0.1.13.0
+class HasGLogFunc env where
+  type GMsg env
+  gLogFuncL :: Lens' env (GLogFunc (GMsg env))
+
+-- | Quick way to run a RIO that only has a logger in its environment.
+--
+-- @since 0.1.13.0
+instance HasGLogFunc (GLogFunc msg) where
+  type GMsg (GLogFunc msg) = msg
+  gLogFuncL = id
+
+-- | A generic logger of some type @msg@.
+--
+-- Your 'GLocFunc' can re-use the existing classical logging framework
+-- of RIO, and/or implement additional transforms,
+-- filters. Alternatively, you may log to a JSON source in a database,
+-- or anywhere else as needed. You can decide how to log levels or
+-- severities based on the constructors in your type. You will
+-- normally determine this in your main app entry point.
+--
+-- @since 0.1.13.0
+newtype GLogFunc msg = GLogFunc (CallStack -> msg -> IO ())
+
+#if MIN_VERSION_base(4,12,0)
+-- https://hackage.haskell.org/package/base-4.12.0.0/docs/Data-Functor-Contravariant.html
+
+-- | Use this instance to wrap sub-loggers via 'RIO.mapRIO'.
+--
+-- The 'Contravariant' class is available in base 4.12.0.
+--
+-- @since 0.1.13.0
+instance Contravariant GLogFunc where
+  contramap = contramapGLogFunc
+  {-# INLINABLE contramap #-}
+#endif
+
+-- | Perform both sets of actions per log entry.
+--
+-- @since 0.1.13.0
+instance Semigroup (GLogFunc msg) where
+  GLogFunc f <> GLogFunc g = GLogFunc (\a b -> f a b *> g a b)
+
+-- | 'mempty' peforms no logging.
+--
+-- @since 0.1.13.0
+instance Monoid (GLogFunc msg) where
+  mempty = mkGLogFunc $ \_ _ -> return ()
+  mappend = (<>)
+
+-- | A vesion of 'contramapMaybeGLogFunc' which supports filering.
+--
+-- @since 0.1.13.0
+contramapMaybeGLogFunc :: (a -> Maybe b) -> GLogFunc b -> GLogFunc a
+contramapMaybeGLogFunc f (GLogFunc io) =
+  GLogFunc (\stack msg -> maybe (pure ()) (io stack) (f msg))
+{-# INLINABLE contramapMaybeGLogFunc #-}
+
+-- | A contramap. Use this to wrap sub-loggers via 'RIO.mapRIO'.
+--
+-- If you are on base > 4.12.0, you can just use 'contramap'.
+--
+-- @since 0.1.13.0
+contramapGLogFunc :: (a -> b) -> GLogFunc b -> GLogFunc a
+contramapGLogFunc f (GLogFunc io) = GLogFunc (\stack msg -> io stack (f msg))
+{-# INLINABLE contramapGLogFunc #-}
+
+-- | Make a custom generic logger. With this you could, for example,
+-- write to a database or a log digestion service. For example:
+--
+-- > mkGLogFunc (\stack msg -> send (Data.Aeson.encode (JsonLog stack msg)))
+--
+-- @since 0.1.13.0
+mkGLogFunc :: (CallStack -> msg -> IO ()) -> GLogFunc msg
+mkGLogFunc = GLogFunc
+
+-- | Log a value generically.
+--
+-- @since 0.1.13.0
+glog ::
+     (MonadIO m, HasCallStack, HasGLogFunc env, MonadReader env m)
+  => GMsg env
+  -> m ()
+glog t = do
+  GLogFunc gLogFunc <- view gLogFuncL
+  liftIO (gLogFunc callStack t)
+{-# INLINABLE glog #-}
+
+--------------------------------------------------------------------------------
+-- Integration with classical logger framework
+
+-- | Level, if any, of your logs. If unknown, use 'LogOther'. Use for
+-- your generic log data types that want to sit inside the classic log
+-- framework.
+--
+-- @since 0.1.13.0
+class HasLogLevel msg where
+  getLogLevel :: msg -> LogLevel
+
+-- | Source of a log. This can be whatever you want. Use for your
+-- generic log data types that want to sit inside the classic log
+-- framework.
+--
+-- @since 0.1.13.0
+class HasLogSource msg where
+  getLogSource :: msg -> LogSource
+
+-- | Make a 'GLogFunc' via classic 'LogFunc'. Use this if you'd like
+-- to log your generic data type via the classic RIO terminal logger.
+--
+-- @since 0.1.13.0
+gLogFuncClassic ::
+     (HasLogLevel msg, HasLogSource msg, Display msg) => LogFunc -> GLogFunc msg
+gLogFuncClassic (LogFunc {unLogFunc = io}) =
+  mkGLogFunc
+    (\theCallStack msg ->
+       liftIO
+         (io theCallStack (getLogSource msg) (getLogLevel msg) (display msg)))
